from django.db import migrations


class Migration(migrations.Migration):
    operations = [
        # There used to be an operation here, which is no longer relevant.
        # We keep the migration to avoid messing with existing deployments
        # of django-tree.
    ]
